import os
import pytest
from spine.utils.env_utils import get_openai_key

def test_get_openai_key_existing_env(tmpdir):
    # Set the environment variable
    os.environ["OPENAI_API_KEY"] = "test-api-key"

    # Test when the environment variable is set
    key = get_openai_key(str(tmpdir.join(".env")))
    assert key == "test-api-key"


def test_get_openai_key_missing_env(tmpdir):
    # Remove the environment variable
    os.environ.pop("OPENAI_API_KEY", None)

    # Test when both the environment variable and .env file are missing
    with pytest.raises(ValueError) as excinfo:
        get_openai_key(str(tmpdir.join(".env")))
    assert str(excinfo.value) == "OPENAI_API_KEY is not set in the .env file or the environment"


def test_get_openai_key_empty_env_file(tmpdir):
    # Create an empty .env file
    tmpdir.join(".env").write("")

    # Test when the .env file is empty
    with pytest.raises(ValueError) as excinfo:
        get_openai_key(str(tmpdir.join(".env")))
    assert str(excinfo.value) == "OPENAI_API_KEY is not set in the .env file or the environment"


if __name__ == "__main__":
    pytest.main([__file__])



"""
Misc utils, mainly for sanity checks and functions to be used within cli.py

Written by: Aaron Ward
"""

import logging
from datetime import date
from pathlib import Path

logger = logging.getLogger('spine.utils.misc')

def read_txt_file(path):
    """
    Reads a text file and returns its content as a string.

    Args:
        path (str): The path to the text file.

    Returns:
        str: The content of the text file as a string.

    Raises:
        FileNotFoundError: If the file specified by the path does not exist.
        IOError: If there was an error reading the file.

    Example:
        >>> content = read_txt_file('path/to/file.txt')
        >>> print(content)
        This is the content of the text file.
    """
    try:
        with open(path, 'r') as file:
            content = file.read()
            return content
    except FileNotFoundError:
        raise FileNotFoundError(f"File '{path}' does not exist.")
    except IOError:
        raise IOError(f"Error reading file '{path}'.")



def check_existence(file_path):
    """
    Function for checking if config file path exists

    args:
        file_path: path of a file or directory
    """

    if not Path(file_path).expanduser().exists():
        logger.error(
            f"Exception: File or directory doesn't exist in: {file_path}\n"
            "Please re-check your config values and try again.")

    return True

def check_config_file(config):
    """
    This function works as a sanity check on the values withing the config
    file. If the required parameters have not been filled out the CLI tool will
    exit and an exception is thrown.

    args:
        config (dict): from config yaml file
    """

    for key in config.keys():
        # If the key is not of dict datatype
        if not isinstance(config[key], dict):
            if config[key] == "":
                logger.error(
                    f"{key} is an empty string. Make sure to populate the config file.")

            # Check path parameters
            # if key == "paths":
            #     if not config['paths']['file'] or not config['paths']['file']:
            #         logger.error(
            #             "You must specify a <some file>")


            # Check document files
            # if key == "reference_files":
            #     for file in config['reference_files'].keys():
            #         if not config['reference_files'][file][-1]:
            #             logger.error(
            #                 "You must specify the suitable reference files")
            #         else:
            #             check_existence(
            #                 Path(
            #                     *config['reference_files'][file]).expanduser())



    logger.info("Config file is valid")


"""
Helper functions for retrieving keys from environment variables.

Written by: Aaron Ward
"""

import os
from dotenv import load_dotenv, dotenv_values

def get_openai_key(env_path):
    """
    Retrieves the OpenAI API key from the environment.

    Args:
        env_path (str): Path to the .env file.
    Returns:
        str: OpenAI API key.
    Raises:
        ValueError: If the OPENAI_API_KEY is not set in the .env file or the environment.
    """

    load_dotenv(dotenv_path=env_path, verbose=True)

    # Set the OPENAI_API_KEY environment variable
    openai_api_key = os.getenv('OPENAI_API_KEY')
    if openai_api_key is None:
        raise ValueError("OPENAI_API_KEY is not set in the .env file or the environment")

    return openai_api_key


import logging
import os
from spine.utils.misc.logging_utils import CustomFormatter, log_to_console, exception, timing

def test_custom_formatter():
    record_info = logging.LogRecord(
        name='test_logger',
        level=logging.INFO,
        pathname='path/to/file',
        lineno=10,
        msg='Test message',
        args=(),
        exc_info=None
    )

    formatter = CustomFormatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    formatted_output = formatter.format(record_info)

    assert 'INFO' in formatted_output
    assert 'Test message' in formatted_output

def test_log_to_console(caplog):
    caplog.set_level(logging.INFO)

    logger_name = 'test_logger'
    is_header = False
    if_n_handlers = 0

    log_to_console(logger_name, is_header, if_n_handlers)

    logger = logging.getLogger(logger_name)
    assert len(logger.handlers) == 2  # StreamHandler and FileHandler

    # Log an INFO message
    logger.info('Test INFO message')
    assert 'Test INFO message' in caplog.text

    # Log an ERROR message
    logger.error('Test ERROR message')
    assert 'Test ERROR message' in caplog.text

def test_exception(caplog):
    caplog.set_level(logging.ERROR)

    logger = logging.getLogger('test_logger')

    @exception(logger)
    def test_function():
        raise ValueError('Test exception')

    try:
        test_function()
    except ValueError:
        pass

    assert 'Exception occured in test_function' in caplog.text
    assert 'ValueError: Test exception' in caplog.text

def test_timing(caplog):
    caplog.set_level(logging.INFO)

    logger = logging.getLogger('test_logger')

    @timing(logger)
    def test_function():
        pass

    test_function()

    assert 'test_function: Start' in caplog.text
    assert 'test_function: End' in caplog.text

if __name__ == "__main__":
    import pytest
    pytest.main()
    
# Delete the logs.log file after all the tests have finished
if os.path.exists('logs.log'):
    os.remove('logs.log')



import logging
import tempfile
from pathlib import Path
from spine.utils.misc.misc_utils import check_existence

def test_check_existence(caplog):
    """
    Unit test for the check_existence function.

    Args:
        caplog: Pytest fixture to capture log messages.

    This test case covers both scenarios: checking the existence of an existing file
    and checking the existence of a non-existent file. It also verifies if the error
    message is logged correctly.

    """
    caplog.set_level(logging.ERROR)

    # Create a temporary directory
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)

        # Create a temporary file
        test_file = temp_path / "test_file.txt"
        test_file.touch()

        # Call the function with an existing file
        assert check_existence(test_file) is True

        # Call the function with a non-existent file
        non_existent_file = temp_path / "non_existent_file.txt"
        assert check_existence(non_existent_file) is True

        # Check the error message logged
        assert "Exception: File or directory doesn't exist" in caplog.text

if __name__ == "__main__":
    import pytest
    pytest.main()



"""
Entry point script for the fleet CLI tool.
Terminal commands and options are specified here.

Written by: Aaron Ward
"""
import ast
import yaml
import click
import logging
import configparser
from pathlib import Path
from pyfiglet import Figlet
from termcolor import colored

from spine import main
from spine.utils.misc import logging_utils
from spine.utils.misc.misc_utils import *

logger = logging.getLogger(__name__)
logging_utils.log_to_console(__name__, is_header=True)
logging_utils.log_to_console('spine.utils.misc')

# for FONTS in `pyfiglet -l`; do echo $FONTS; pyfiglet $FONTS -f $FONTS; done;
print(colored(Figlet(font="univers").renderText("Spine"), "magenta"))


############### CLI Commands ###############
@click.group()
def cli():
    """
    \033[0;35m
    * The backbone for your agents 🦴 \n
    An Autonomous LLM powered scheduler and agent orchestration system
     \033[0;0m
    """
    pass


@cli.command()
@click.option('--config-path',
              help='Path to config file containing parameters and modelling templates, copy from config_template.yml',
              required=True)
def run(config_path):
    """
    Run the spine server

    \033[0;35mexample: spine run --config-path="./config.yml"\033[0;0m
    """
    logger.info("Running spine run command")
    check_existence(config_path)
    config = configparser.RawConfigParser()
    config.read(config_path)
        
    # TODO: Update with config validation rather than yaml/dict
    # https://docs.python.org/3/library/configparser.html#
    config_dict = dict(config.items())
    check_config_file(config_dict)

    # run the pipeline
    main.run(config)
    


"""
This is the main driver program for spine. 

Written by: Aaron Ward
"""
import os
import json
import logging
import configparser
from pathlib import Path

from spine.utils.misc import (logging_utils, misc_utils)
from spine.utils.env_utils import get_openai_key

logger = logging.getLogger(__name__)
logging_utils.log_to_console(__name__, is_header=True)

def run(config):
    """
    This is the main run function of the spine application. 
    It spins up the spine server, appropraite nodes and 
    the chat interface.

    Args:
        - config (configparser): config passed to spine run
    Returns:
        - None
    """

if __name__ == "__main__":
    pass



"""
Logger utilities for the spine package.
"""
import logging
import functools
from functools import wraps
from time import time

class CustomFormatter(logging.Formatter):
    """
    Logging colored formatter, adapted from https://stackoverflow.com/a/56944256/3638629

    This CustomFormatter is set with is_header=True, otherwise general information
    should be logged in standard white text.
    """

    red = "\x1b[31;20m"
    magenta = '\x1b[0;35m'
    reset = '\x1b[0m'

    def __init__(self, fmt):
        super().__init__()
        self.fmt = fmt
        self.FORMATS = {
            logging.INFO: self.magenta + self.fmt + self.reset,
            logging.ERROR: self.red + self.fmt + self.reset,
        }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def log_to_console(logger_name, is_header=False, if_n_handlers = 0):
    '''
    Set logger configurations.

    args:
        logger_name (str): name of the logger
        is_header (bool): whether to log info in magenta
        if_n_handlers (int): number of handlers to add to the logger
    '''
    logger = logging.getLogger(logger_name)

    if len(logger.handlers) == if_n_handlers:
        logger.setLevel(logging.INFO)

        # Make stream handler to log to console
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)

        # Make file handler to log to file
        fh = logging.FileHandler("logs.log")
        fh.setLevel(logging.INFO)

        # Set format of log messages
        formatter = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        fh.setFormatter(logging.Formatter(formatter))

        if is_header:
            ch.setFormatter(CustomFormatter(formatter))
        else:
            ch.setFormatter(logging.Formatter(formatter))

        # Add handlers to logger
        logger.addHandler(ch)
        logger.addHandler(fh)


def exception(logger):
    '''
    returns decorator logging exceptions
    of the wrapped function with the given logger.

    args:
      logger (logging.Logger): a logger
    returns:
        decorator (function): a decorator function
    '''
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as err:
                logger.exception(f'Exception occured in {func.__name__}')
                raise
        return wrapper
    return decorator

def timing(logger):
    '''
    returns decorator that logs the total execution time
    of the wrapped function with the given logger.

    args:
        logger (logging.Logger): a logger
    returns:
        decorator (function): a dsecorator function
    '''
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            logger.info('{}: Start'.format(func.__name__))
            start = time()
            result = func(*args, **kwargs)
            duration = time()-start
            if duration > 3600:
                duration = duration/3600
                units = 'hrs'
            elif duration > 60:
                duration = duration/60
                units = 'min'
            else:
                units = 'sec'
            logger.info('{}: End ({:.4f} {})'.format(func.__name__, duration, units))
            return result
        return wrapper
    return decorator